# from .db import Database
from collections.abc import MutableSet
import mcol
import mrich


class TagTable:
    """Object representing the 'tag' table in the :class:`.Database`.

    .. attention::

            :class:`.TagTable` objects should not be created directly. Instead use the :meth:`.HIPPO.tags` property.

    """

    _table = "tag"

    def __init__(
        self,
        db: "Database",
    ) -> None:

        self._db = db

    ### FACTORIES

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def table(self) -> str:
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def unique(self) -> set[str]:
        """Returns a set of unique tag names contained in the table"""
        values = self.db.select(
            table=self.table, query="DISTINCT tag_name", multiple=True
        )
        return set(v for v, in values)

    ### METHODS

    def summary(self) -> "pd.DataFrame":
        """Print a summary table of tags with compound and pose counts"""

        from pandas import DataFrame

        sql = """
        SELECT tag_name,
        COUNT(DISTINCT tag_compound),
        COUNT(DISTINCT tag_pose)
        FROM tag
        GROUP BY tag_name
        ORDER BY tag_name;
        """

        cursor = self.db.execute(sql)

        data = [
            dict(tag=a, num_compounds=b, num_poses=c) for a, b, c in cursor.fetchall()
        ]

        df = DataFrame(data)
        df = df.set_index("tag")

        # compounds with poses

        sql = """
        SELECT tag_name, COUNT(DISTINCT pose_compound) FROM tag
        INNER JOIN pose
        ON tag_pose = pose_id
        GROUP BY tag_name
        ORDER BY tag_name;
        """

        cursor = self.db.execute(sql)

        for tag, count in cursor.fetchall():
            df.loc[tag, "num_posed_compounds"] = count

        df = df.astype(int)

        mrich.print(df)

        return df

    def rename(self, old: str, new: str) -> None:
        """Rename all instances of a tag across the database"""

        sql = """
        UPDATE OR IGNORE tag
        SET tag_name = ?2
        WHERE tag_name = ?1;
        """

        self.db.execute(sql, (str(old), str(new)))

        self.delete(old)

        self.db.commit()

    def delete(self, tag: str) -> None:
        """Delete all assignments for the given tag"""
        self.db.delete_where(table="tag", key="name", value=tag)

    ### DUNDERS

    def __str__(self) -> str:
        """Unformatted representation of this object"""
        return f"Tags {self.unique}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"


class TagSet(MutableSet):
    """Object representing a subset of the 'tag' table in the :class:`.Database` belonging to a certain :class:`.Compound` or :class:`.Pose`.

    .. attention::

            :class:`.TagSet` objects should not be created directly. Instead use the :meth:`.Compound.tags` or :meth:`.Pose.tags` property.

    """

    def __init__(
        self,
        parent: "Compound | Pose",
        tags: list | tuple | None = None,
        immutable: bool = False,
        commit: bool = True,
    ):

        self._elements = []
        self._immutable = immutable
        self._parent = parent

        tags = tags or []

        for tag in tags:
            if tag not in self._elements:
                self._elements.append(tag)

        # for tag in tags:
        #     self.add(tag, commit=False)

        # if commit:
        #     self.db.commit()

    ### FACTORIES

    ### PROPERTIES

    @property
    def tags(self) -> list:
        """Returns the elements in this set"""
        return self._elements

    @property
    def immutable(self) -> bool:
        """Is this set is immutable?"""
        return self._immutable

    @immutable.setter
    def immutable(
        self,
        b: bool,
    ) -> None:
        self._immutable = b

    @property
    def parent(self):
        """Returns this set of tags parent :class:`.Compound` or :class:`.Pose`."""
        return self._parent

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self.parent.db

    ### DATABASE

    def _remove_tag_from_db(
        self,
        tag: str,
    ) -> None:
        """Delete a specific tag assignment for the parent :class:`.Compound`/:class:`.Pose`

        :param tag: tag to delete

        """
        sql = f'DELETE FROM tag WHERE tag_name="{tag}" AND tag_{self.parent.table} = {self.parent.id}'
        self.db.execute(sql)

    def _clear_tags_from_db(
        self,
        tag: str,
    ) -> None:
        """Delete all tag assignments for the parent :class:`.Compound`/:class:`.Pose`

        :param tag: tag to delete

        """
        sql = f"DELETE FROM tag WHERE tag_{self.parent.table} = {self.parent.id}"
        self.db.execute(sql)

    def _add_tag_to_db(
        self,
        tag: str,
        commit: bool = True,
    ) -> None:
        """Assign a given tag to the parent

        :param tag: tag to add
        :param commit: commit the changes? (Default value = True)

        """
        payload = {"name": tag, self.parent.table: self.parent.id}
        self.db.insert_tag(**payload, commit=commit)

    ### METHODS

    def pop(self) -> str:
        """Pop the last element"""
        assert not self.immutable
        return self._elements.pop()

    def discard(
        self,
        tag: str,
    ) -> None:
        """Discard an element

        :param tag: tag to discard

        """

        self.discard(tag)

    def clear(self):
        """Clear all tags"""
        self._elements = []
        self._clear_tags_from_db(self)

    def remove(self, tag: str) -> None:
        """Remove an element

        :param tag: tag to remove
        :raises ValueError: if tag is not in set

        """

        assert not self.immutable
        if tag in self:
            i = self._elements.index(tag)
            del self._elements[i]
            self._remove_tag_from_db(tag)
        else:
            raise ValueError(f"{tag} not in {self}")

    def add(
        self,
        tag: str,
        commit: bool = True,
    ) -> None:
        """Add a tag to the set

        :param tag: tag to add
        :param commit: commit the change? (Default value = True)

        """

        assert not self.immutable
        if tag not in self._elements:
            self._elements.append(tag)
            self._add_tag_to_db(tag, commit=commit)

    def glob(self, pattern: str) -> list[str]:
        """Construct a list from tags in the set names that match a given UNIX-style pattern.

        :param pattern: unix style pattern with shell-style wildcards
        :returns: list of tags

        """

        import fnmatch

        return fnmatch.filter(self.tags, pattern)

    ### DUNDERS

    def __contains__(self, tag: str) -> bool:
        """Is this tag in the set?"""
        return tag in self.tags

    def __str__(self) -> str:
        """Unformatted representation of this object"""
        return str(self._elements)

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"

    def __len__(self) -> int:
        """Number of tags in this set"""
        return len(self._elements)

    def __iter__(self):
        """Iterate through this set"""
        return iter(self._elements)

    def __add__(self, other):
        """
        .. attention::

                Adding sets together is not supported

        """
        raise NotImplementedError

    def __getitem__(self, key: int):
        """Get a specific element in the set by index"""
        return self._elements[key]



import apsw


def executemany(path, sql, payload):
    connection = apsw.Connection(str(path.resolve()))
    result = list(connection.executemany(sql, payload))
    connection.close()
    return result



import mrich


mrich.debug("importing posebusters.PoseBusters")
from posebusters import PoseBusters

# setup posebusters
mrich.debug("Setting up PoseBusters(config='mol')")
MOL_ENGINE = PoseBusters(config="mol")
MOL_ENGINE._initialize_modules()
MOL_ENGINE.module_name.pop(0)
MOL_ENGINE.module_func.pop(0)
MOL_ENGINE.module_args.pop(0)
MOL_ENGINE.fname.pop(0)
MOL_ENGINE.module_test_names = {
    c["name"]: c["chosen_binary_test_output"] for c in MOL_ENGINE.config["modules"]
}


def check_mol(mol, debug: bool = False):

    failed = False

    for name, fname, func, args in zip(
        MOL_ENGINE.module_name,
        MOL_ENGINE.fname,
        MOL_ENGINE.module_func,
        MOL_ENGINE.module_args,
    ):

        assert len(args) == 1 and "mol_pred" in args

        results = func(mol_pred=mol)

        test_names = MOL_ENGINE.module_test_names[name]

        for test_name in test_names:

            passed = results["results"][test_name]

            if debug:
                mrich.var(f"{name}.{test_name}", results["results"][test_name])

            if not passed:
                failed = True
                mrich.error(f"Molecule failed {name}.{test_name}")

    return not failed



from pathlib import Path
import shutil

import logging

logging.getLogger("PIL").setLevel(logging.WARNING)

import mrich


class ProjectPage:
    """
    Recipe proposal web page

    """

    def __init__(
        self,
        output_dir: str | Path,
        *,
        animal: "HIPPO",
        scaffolds: "CompoundSet | None" = None,
        suppliers: list[str] | None = None,
        starting_recipe: "Recipe | None" = None,
        rgen: "RandomRecipeGenerator | None" = None,
        scorer: "Scorer | None" = None,
        proposals: "list[Recipe] | None" = None,
        title: str | None = None,
        scaffold_tag: str = "Syndirella base",
        extra_recipe_dir: str | Path = None,
        skip_existing: bool = True,
    ):

        # setup directories
        self._output_dir = Path(output_dir)
        self.make_directories()

        # project objects
        self._animal = animal
        self._scaffolds = scaffolds
        self._suppliers = suppliers
        self._starting_recipe = starting_recipe
        self._rgen = rgen
        self._scorer = scorer
        self._proposals = proposals
        self._extra_recipe_dir = extra_recipe_dir

        self._all_scaffolds = self.animal.compounds(tag=scaffold_tag)

        mrich.debug(f"{len(self.all_scaffolds)=}")

        self._all_scaffold_poses = None
        self._all_elabs = None
        self._all_elab_poses = None
        self._scaffold_poses = None

        self._title = title or animal.name

        self._skip_existing = skip_existing

        self.setup_page()
        self.write_html()

    ### FACTORIES

    ### PROPERTIES

    @property
    def animal(self) -> "HIPPO":
        """associated :class:`.HIPPO` object"""
        return self._animal

    @property
    def db(self) -> "Database":
        """associated :class:`.Database` object"""
        return self.animal.db

    @property
    def doc(self) -> "yattag.Doc":
        """yattag.Doc"""
        return self._doc

    @property
    def tag(self) -> "yattag.tag":
        """yattag.tag"""
        return self._tag

    @property
    def text(self) -> "yattag.text":
        """yattag.text"""
        return self._text

    @property
    def line(self) -> "yattag.line":
        """yattag.line"""
        return self._line

    @property
    def title(self) -> str:
        """Page title"""
        return self._title

    @property
    def output_dir(self) -> "Path":
        """Output directory"""
        return self._output_dir

    @property
    def resource_dir(self) -> "Path":
        """Output directory"""
        return self.output_dir / "web_resources"

    @property
    def mol_image_dir(self) -> "Path":
        """Output directory"""
        return self.resource_dir / "mol_images"

    @property
    def pose_sdf_dir(self) -> "Path":
        """Output directory"""
        return self.resource_dir / "pose_sdfs"

    @property
    def index_path(self) -> "Path":
        """index.html Path"""
        return self.output_dir / "index.html"

    @property
    def proposals(self):
        return self._proposals

    @property
    def scaffolds(self):
        return self._scaffolds

    @property
    def suppliers(self):
        return self._suppliers

    @property
    def starting_recipe(self):
        return self._starting_recipe

    @property
    def rgen(self):
        return self._rgen

    @property
    def scorer(self):
        return self._scorer

    @property
    def all_scaffolds(self):
        return self._all_scaffolds

    @property
    def all_elabs(self):
        if self._all_elabs is None:
            self._all_elabs = self.all_scaffolds.elabs
        return self._all_elabs

    @property
    def all_elab_poses(self):
        if self._all_elab_poses is None:
            self._all_elab_poses = self.all_elabs.poses
        return self._all_elab_poses

    @property
    def scaffold_poses(self):
        if self._scaffold_poses is None:
            self._scaffold_poses = self.scaffolds.poses
        return self._scaffold_poses

    @property
    def all_scaffold_poses(self):
        if self._all_scaffold_poses is None:
            self._all_scaffold_poses = self.all_scaffolds.poses
        return self._all_scaffold_poses

    @property
    def proposal(self):
        if len(self.proposals) != 1:
            mrich.warning(f"{len(self.proposals)=}")
        return self._proposals[0]

    @property
    def extra_recipe_dir(self):
        return self._extra_recipe_dir

    @property
    def skip_existing(self) -> bool:
        return self._skip_existing

    ### METHODS

    def write_html(self) -> None:
        """Write the index.html file"""

        from yattag import indent

        path = self.index_path

        with open(path, "wt") as f:
            mrich.writing(path)
            f.writelines(indent(self.doc.getvalue()))

    ### INTERNAL HTML STUFF

    def make_directories(self) -> None:
        """Create output directories"""

        mrich.writing(self.output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        mrich.writing(self.resource_dir)
        (self.resource_dir).mkdir(parents=True, exist_ok=True)

        mrich.writing(self.mol_image_dir)
        (self.mol_image_dir).mkdir(parents=True, exist_ok=True)

        mrich.writing(self.pose_sdf_dir)
        (self.pose_sdf_dir).mkdir(parents=True, exist_ok=True)

    def setup_page(self) -> None:
        """Create the yattag page content"""

        # yattag setup
        from yattag import Doc

        doc, tag, text, line = Doc().ttl()

        self._doc = doc
        self._tag = tag
        self._text = text
        self._line = line

        self.doc.asis("<!DOCTYPE html>")

        with self.tag("html"):

            self.header()

            with self.tag("body", klass="w3-content", style="max-width:none"):

                with self.tag("div", klass="w3-bar w3-teal"):
                    with self.tag("div", klass="w3-bar-item"):
                        src = "https://github.com/mwinokan/HIPPO/raw/main/logos/hippo_assets-02.png?raw=true"
                        self.doc.stag(
                            "img", src=src, style="max-height:75px"
                        )  # , klass="w3-image")

                    with self.tag("div", klass="w3-bar-item"):
                        with self.tag("h1"):
                            self.text(self.title)

                with self.tag("div", klass="w3-container w3-dark-gray w3-padding"):
                    self.section(self.sec_targets)
                    self.section(self.sec_hits)

                    # placeholders
                    if self.scaffolds:
                        self.section(self.sec_scaffolds)
                    if self.scaffolds:
                        self.section(self.sec_elaborations)
                    # if self.quoting: self.section(self.sec_quoting)
                    # if self.product_pool: self.section(self.sec_product_pool)
                    # if self.route_pool: self.section(self.sec_route_pool)
                    if self.rgen:
                        self.section(self.sec_rgen)
                    if self.scorer:
                        self.section(self.sec_scorer)
                    if self.proposals:
                        self.section(self.sec_proposals)

                with self.tag("div", klass="w3-container w3-teal w3-padding"):
                    with self.tag("div", klass="w3-center"):
                        src = "https://github.com/mwinokan/HIPPO/raw/main/logos/hippo_logo_tightcrop.png?raw=true"
                        self.doc.stag("img", src=src, style="max-height:150px")

    def header(self) -> None:
        """Create the page header"""

        with self.tag("head"):

            with self.tag("title"):
                self.text(self.title)

            self.doc.stag("meta", charset="UTF-8")
            self.doc.stag(
                "meta", name="viewport", content="width=device-width, initial-scale=1"
            )
            self.doc.stag(
                "link",
                rel="stylesheet",
                href="https://www.w3schools.com/w3css/4/w3.css",
            )
            self.doc.stag(
                "link",
                rel="stylesheet",
                href="https://fonts.googleapis.com/css?family=Oswald",
            )
            self.doc.stag(
                "link",
                rel="stylesheet",
                href="https://fonts.googleapis.com/css?family=Open Sans",
            )
            self.doc.stag(
                "link",
                rel="stylesheet",
                href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css",
            )

            with self.tag("script", src="https://cdn.plot.ly/plotly-latest.min.js"):
                ...

            with self.tag("script", src="https://3Dmol.org/build/3Dmol-min.js"):
                ...

            with self.tag("script", src="https://3Dmol.org/build/3Dmol.ui-min.js"):
                ...

            self.style()

    def style(self) -> None:
        """Create the page style"""

        # change to a .css file and use doc.stag("link", rel="stylesheet", href="style.css")

        with self.tag("style"):
            self.doc.asis(
                """h1,h2,h3,h4,h5,h6 {font-family: "Oswald"}body {font-family: "Open Sans"}"""
            )

    def section_header(self, title: str, tag: str = "h2") -> None:
        """section header"""
        with self.tag(tag):
            self.text(str(title))

    def accordion(self) -> None:
        """sub-content accordion"""
        # https://www.w3schools.com/w3css/w3css_accordions.asp
        raise NotImplementedError

    def sidebar(self) -> None:
        # https://www.w3schools.com/w3css/w3css_sidebar.asp
        raise NotImplementedError

    def var(self, key, value, tag=None, separator=": ") -> None:
        """sub-content accordion"""
        text = f"{key}{separator}{value}"

        if not tag:
            with self.tag("b"):
                self.text(key)
            self.text(separator)
            self.text(str(value))
        else:
            with self.tag(tag):
                self.var(key, value, separator=separator)

    def section(self, function) -> None:
        """create section div"""
        with self.tag("div", klass="w3-panel w3-border w3-white"):
            function()

    def plotly_graph(self, figure, filename, write: bool = True):

        # from plotly.offline import plot
        from hippo_plot import write_html

        f"""<div> 
            <iframe src="iter1b_go/scoring_newfeat_coverage.html" width="100%" height="500" style="border:none;"></iframe>
        </div>"""

        path = self.resource_dir / filename
        rel_path = Path(self.resource_dir.name) / filename

        if write:
            write_html(path, figure)

        # embed the graph
        with self.tag("div"):
            with self.tag(
                "iframe",
                src=str(rel_path),
                width="100%",
                height="500",
                style="border:none",
            ):
                ...

    def table(self, data, style: str = "w3-table-all w3-responsive", **kwargs):
        """Embed some data as a table"""
        from pandas import DataFrame

        df = DataFrame(data)
        html = df.to_html(**kwargs, classes=style, index=False, escape=False)
        self.doc.asis(html)
        self.doc.asis("<br>")

    # def mol_grid_svg(self, cset, **kwargs):

    #     mols = [c.mol for c in cset]

    #     from rdkit.Chem.Draw import MolsToGridImage

    #     return MolsToGridImage(mols,
    #         molsPerRow=3,
    #         subImgSize=(200,
    #         200),
    #         legends=None,
    #         # highlightAtomLists=None,
    #         # highlightBondLists=None,
    #         useSVG=True,
    #         returnPNG=False,
    #         **kwargs)

    def save_compound_image(self, compound):
        from rdkit.Chem.Draw import MolToImage

        image = MolToImage(compound.mol)
        path = self.mol_image_dir / f"C{compound.id}.png"
        mrich.writing(path)
        image.save(path)

    def save_pose_sdf(self, pose):
        path = self.pose_sdf_dir / f"P{pose.id}.sdf"
        self.animal.poses([pose.id]).write_sdf(path, inspirations=False)

    def save_pset_sdf(self, name, pset):
        path = self.pose_sdf_dir / f"{name}.sdf"
        pset.write_sdf(path, inspirations=False)

    def compound_image(self, compound, max_height="250px"):
        self.save_compound_image(compound)

        src = str(
            Path(self.resource_dir.name)
            / Path(self.mol_image_dir.name)
            / f"C{compound.id}.png"
        )

        self.doc.stag("img", src=src, style=f"max-height:{max_height}")

    def pose_3d_view(self, pose):

        self.save_pose_sdf(compound)

        src = str(
            Path(self.resource_dir.name)
            / Path(self.mol_image_dir.name)
            / f"C{compound.id}.png"
        )

        self.doc.stag("img", src=src, style=f"max-height:{max_height}")

    def compound_grid(self, compounds, style="w3-center", pose_modal: bool = False):

        id_num_poses_dict = compounds.id_num_poses_dict

        inspiration_map = self.db.get_compound_id_inspiration_ids_dict()

        with self.tag("div", klass="w3-row"):
            for compound in compounds:

                with self.tag(
                    "div",
                    klass=f"w3-col s12 m6 l4 {style} w3-hover-border-black",
                    style="border:8px solid white",
                ):
                    with self.tag("p"):
                        with self.tag("b"):
                            self.text(f"{compound}")

                    self.compound_image(compound)

                    with self.tag("p", klass="w3-small w3-monospace"):
                        self.text(f"{compound.inchikey}")
                        self.doc.asis("<br>")
                        self.text(f"{compound.smiles}")
                        self.doc.asis("<br>")

                        inspirations = inspiration_map.get(compound.id, None)

                        if (
                            not inspirations
                            and "inspiration_pose_ids" in compound.metadata
                        ):
                            inspirations = compound.metadata["inspiration_pose_ids"]

                        if inspirations:
                            inspirations = self.animal.poses[inspirations]
                            self.text(f"inspirations: {inspirations.names}")
                        else:
                            self.text(f"inspirations: ?")

                    num_poses = id_num_poses_dict[compound.id]
                    self.button(f"{num_poses} poses", disable=num_poses == 0)

                    if pose_modal:
                        poses = compound.poses

                        modal_name = f"modal_c{compound.id}_poses"

                        # MOLECULE MODAL
                        self.modal_button(
                            f"view {len(poses)} poses",
                            modal_name,
                            disable=len(poses) == 0,
                        )

                        if poses:

                            self.save_pset_sdf(modal_name, poses)

                            def modal_content():
                                with self.tag("p"):
                                    self.text("TEXT TEXT TEXT")

                            self.modal(modal_name, modal_content)

    def modal_button(
        self, text, modal_name, disable: bool = False, style: str = "w3-teal"
    ):
        onclick = f"document.getElementById('{modal_name}').style.display='block'"
        self.button(text, style=style, onclick=onclick, disable=disable)

    def button(
        self, text: str, onclick: str = "", style="w3-teal", disable: bool = False
    ):

        assert text

        klass = f"w3-btn {style}"

        if disable:
            klass += " w3-disabled"
            onclick = ""

        with self.tag("button", klass=klass, onclick=onclick):
            self.text(text)

    def modal(self, modal_name, content_function):
        with self.tag("div", id=modal_name, klass="w3-modal"):
            with self.tag("div", klass="w3-modal-content"):
                with self.tag("div", klass="w3-container"):
                    with self.tag(
                        "span",
                        onclick=f"document.getElementById('{modal_name}').style.display='none'",
                        klass="w3-button w3-display-topright",
                    ):
                        self.doc.asis("&times")
                    content_function()

    def recipe_subsection(
        self, recipe, title, sankey: bool = False, title_style="h3", show_title=True
    ):

        if show_title:
            self.section_header(title, title_style)

        recipe_name = title.lower().replace(" ", "_")

        if sankey:
            fig = recipe.sankey()
            self.plotly_graph(fig, f"{recipe_name}.html")

        # self.section_header("Products", "h4")

        df = recipe.products.compounds.get_df()

        def modal_content():
            self.table(df, style="w3-table-all w3-small")

        modal_name = f"{recipe_name}_products"
        self.modal_button("products", modal_name)
        self.modal(modal_name, modal_content)

        if intermediates := recipe.intermediates:
            # self.section_header("Intermediates", "h4")
            df = intermediates.compounds.get_df()

            def modal_content():
                self.table(df, style="w3-table-all w3-small")

            modal_name = f"{recipe_name}_intermediates"
            self.modal_button("intermediates", modal_name)
            self.modal(modal_name, modal_content)

        # self.section_header("Reactants", "h4")

        df = recipe.reactants.df

        def modal_content():
            self.table(df, style="w3-table-all w3-small")

        modal_name = f"{recipe_name}_reactants"
        self.modal_button("reactants", modal_name)
        self.modal(modal_name, modal_content)

        # self.section_header("Reactions", "h4")

        df = recipe.reactions.get_df(mols=False)

        def modal_content():
            self.table(df, style="w3-table-all w3-small")

        modal_name = f"{recipe_name}_reactions"
        self.modal_button("reactions", modal_name)
        self.modal(modal_name, modal_content)

    def scorer_attribute(self, attribute, histogram: bool = True):

        from .scoring import DEFAULT_ATTRIBUTES

        key = attribute.key

        self.section_header(f'{attribute._type}: "{key}"')

        with self.tag("ul"):
            self.var("weight", f"{attribute.weight:.2f}", tag="li")
            self.var("inverse", f"{attribute.inverse}", tag="li")
            self.var("min", f"{attribute.min:.2f}", tag="li")
            self.var("max", f"{attribute.max:.2f}", tag="li")
            self.var("mean", f"{attribute.mean:.2f}", tag="li")
            self.var("std", f"{attribute.std:.2f}", tag="li")

            if key in DEFAULT_ATTRIBUTES:
                description = DEFAULT_ATTRIBUTES[key]["description"]
                if attribute.inverse:
                    description += "(Lower is better)"
                else:
                    description += "(Lower is better)"

                self.var("Description", description, tag="li")

        if histogram:
            fig = attribute.histogram(progress=True)
            self.plotly_graph(fig, f"attribute_{key}_hist.html")

    ### SECTION CONTENT

    def sec_targets(self) -> None:
        """Section on targets"""

        title = "Protein Target"
        targets = self.animal.targets

        if len(targets) > 1:
            title += "s"

        self.section_header(title)

        for target in targets:
            self.section_header(target.name, "h3")

            self.var("name", target.name)

            subsites = target.subsites

            if subsites:

                self.section_header("Subsites", "h4")
                with self.tag("ul"):
                    for subsite in subsites:
                        self.var(f"Site {subsite.id}", subsite.name, tag="li")

        # try:
        fig = self.funnel()
        self.plotly_graph(fig, "project_funnel.html")
        # except Exception as e:
        # mrich.error(e)

    def sec_hits(self) -> None:
        """Section on experimental hits"""

        title = "Experimental hits"
        hit_compounds = self.animal.compounds(tag="hits")
        hit_poses = self.animal.poses(tag="hits")

        self.section_header(title)

        with self.tag("ul"):
            self.var("#compounds", len(hit_compounds), tag="li")
            self.var("#observations", len(hit_poses), tag="li")

        from .animal import GENERATED_TAG_COLS

        # tag statistics
        fig = self.animal.plot_tag_statistics(
            show_compounds=False,
            poses=hit_poses,
            logo=None,
            title="Tags",
            skip=["Pose", "hits"] + GENERATED_TAG_COLS,
        )

        self.plotly_graph(fig, "hit_tags.html")

        # files
        self.section_header("Downloads", "h3")
        path = self.resource_dir / "hit_poses.sdf"
        rel_path = Path(self.resource_dir.name) / "hit_poses.sdf"
        hit_poses.write_sdf(path, inspirations=False)
        table_data = [
            dict(
                Name="hit_poses.sdf",
                Description="SDF of the experimental hits",
                Download=f'<a href="{rel_path}" download>SDF</a>',
            )
        ]
        self.table(table_data)

    def sec_scaffolds(self) -> None:
        """Section on scaffolds"""

        title = "Scaffolds"
        self.section_header(title)

        self.section_header("All scaffolds", "h3")

        with self.tag("ul"):
            self.var("#compounds", len(self.all_scaffolds), tag="li")

        # route dict?
        df = self.all_scaffolds.get_df(mol=False, num_poses=True)  # , routes=True)

        def modal_content():
            self.table(df, style="w3-table-all w3-small")

        modal_name = "all_scaffolds_modal"
        self.modal_button(f"all scaffolds table", modal_name)
        self.modal(modal_name, modal_content)

        # quoting?

        self.section_header("Selected scaffolds", "h3")

        self.compound_grid(self.scaffolds, pose_modal=False)

        # files
        self.section_header("Downloads", "h3")

        table_data = []

        path = self.resource_dir / "all_scaffold_smiles.csv"
        rel_path = Path(self.resource_dir.name) / path.name
        self.scaffolds.write_smiles_csv(path)
        table_data.append(
            dict(
                Name="All scaffolds",
                Description="CSV of scaffold SMILES",
                Download=f'<a href="{rel_path}" download>CSV</a>',
            )
        )

        path = self.resource_dir / "selected_scaffold_smiles.csv"
        rel_path = Path(self.resource_dir.name) / path.name
        self.scaffolds.write_smiles_csv(path)
        table_data.append(
            dict(
                Name="Selected scaffolds",
                Description="CSV of scaffold SMILES",
                Download=f'<a href="{rel_path}" download>CSV</a>',
            )
        )

        self.table(table_data)

    def sec_elaborations(self) -> None:
        """Section on elaborations"""

        elabs = self.scaffolds.elabs

        if not elabs:
            mrich.warning("No elaborations")
            return None

        self._elaborations = elabs

        title = "Elaborations"
        self.section_header(title)

        fig = self.animal.plot_reaction_funnel(
            title="Syndirella elaboration space", logo=False
        )
        self.plotly_graph(fig, "reaction_funnel.html")

    def sec_quoting(self) -> None:
        """Section on quoting"""

        title = "Quoting"
        self.section_header(title)

    def sec_product_pool(self) -> None:
        """Section on product_pool"""

        title = "product_pool"
        self.section_header(title)

    def sec_route_pool(self) -> None:
        """Section on route_pool"""

        title = "route_pool"
        self.section_header(title)

    def sec_rgen(self) -> None:
        """Section on rgen"""

        title = "Random Recipe Generation"
        self.section_header(title)

        rgen = self.rgen

        with self.tag("ul"):
            self.var("suppliers", rgen.suppliers, tag="li")
            self.var("max_lead_time", rgen.max_lead_time, tag="li")
            self.var("route_pool", len(rgen.route_pool), tag="li")

        self.recipe_subsection(rgen.starting_recipe, "Starting Recipe", sankey=False)

    def sec_scorer(self) -> None:
        """Section on scorer"""

        title = "Recipe Selection"
        self.section_header(title)

        scorer = self.scorer

        with self.tag("ul"):
            self.var("#recipes", len(scorer.recipes), tag="li")
            self.var("#attributes", len(scorer.attributes), tag="li")

        fig = scorer.plot(["price", "score"])
        self.plotly_graph(fig, "scorer_scatter.html")

        for attribute in scorer.attributes:
            self.scorer_attribute(attribute)

    def sec_proposals(self) -> None:
        """Section on proposals"""

        title = "Proposal Recipes"
        self.section_header(title)

        table_data = []

        for proposal in self.proposals:

            d = {}

            d["hash"] = str(proposal.hash)
            d["price"] = str(proposal.price)
            d["price/compound"] = str(proposal.price / proposal.num_products)

            for attribute in self.scorer.attributes:
                d[f"{attribute.key} w={attribute.weight}"] = (
                    f"{attribute.get_value(proposal):.1f} ({attribute.unweighted(proposal):.0%})"
                )

            table_data.append(d)

        self.table(table_data)

        for proposal in self.proposals:

            filename = f"proposal_{proposal.hash}.html"
            path = self.resource_dir / filename

            self.section_header(str(proposal), "h3")

            from .plotting import plot_compound_tsnee

            if path.exists() and self.skip_existing:

                self.plotly_graph(None, filename, write=False)

            else:

                fig = plot_compound_tsnee(
                    proposal.products.compounds,
                    logo=False,
                    legend=False,
                    title="Product Clustering",
                )

                self.plotly_graph(fig, filename)

            self.recipe_subsection(
                proposal, f"Recipe {proposal.hash}", sankey=False, show_title=False
            )

        # files
        self.section_header("Downloads", "h3")

        table_data = []

        for proposal in self.proposals:

            # JSON
            filename = f"Recipe_{proposal.hash}.json"
            original = self.rgen.recipe_dir / filename
            if not original.exists() and self.extra_recipe_dir:
                original = Path(self.extra_recipe_dir) / filename
            shutil.copyfile(original, self.resource_dir / filename)

            path = self.resource_dir / filename
            rel_path = Path(self.resource_dir.name) / filename

            table_data.append(
                dict(
                    Name=str(proposal),
                    Description="Recipe JSON",
                    Download=f'<a href="{rel_path}" download>JSON</a>',
                )
            )

            # SDF
            filename = f"Recipe_{proposal.hash}_poses.sdf"

            try:
                original = self.rgen.recipe_dir / filename
                if not original.exists() and self.extra_recipe_dir:
                    original = Path(self.extra_recipe_dir) / filename
                shutil.copyfile(original, self.resource_dir / filename)

                path = self.resource_dir / filename
                rel_path = Path(self.resource_dir.name) / filename

                table_data.append(
                    dict(
                        Name=str(proposal),
                        Description="Recipe product poses (Fragalysis compatible)",
                        Download=f'<a href="{rel_path}" download>SDF</a>',
                    )
                )
            except FileNotFoundError:
                # hit_poses.write_sdf(path, inspirations=False)
                mrich.error(f"Could not find pose SDF: {original}")

            # CAR CSVs

            filename = f"Recipe_{proposal.hash}_CAR"
            path = self.resource_dir / f"{filename}.csv"
            proposal.write_CAR_csv(path)

            for file in Path(self.resource_dir).glob(f"{filename}*.csv"):

                rel_path = Path(self.resource_dir.name) / file.name

                table_data.append(
                    dict(
                        Name=str(proposal),
                        Description=f"CAR input file [{file.name}]",
                        Download=f'<a href="{rel_path}" download>CSV</a>',
                    )
                )

            # Reactant CSV

            filename = f"Recipe_{proposal.hash}_reactants"
            path = self.resource_dir / f"{filename}.csv"

            if not path.exists() or not self.skip_existing:
                proposal.write_reactant_csv(path)

            rel_path = Path(self.resource_dir.name) / path.name

            table_data.append(
                dict(
                    Name=str(proposal),
                    Description=f"Reactant data file",
                    Download=f'<a href="{rel_path}" download>CSV</a>',
                )
            )

            # Product CSV

            filename = f"Recipe_{proposal.hash}_products"
            path = self.resource_dir / f"{filename}.csv"

            if not path.exists() or not self.skip_existing:
                proposal.write_product_csv(path)

            rel_path = Path(self.resource_dir.name) / path.name

            table_data.append(
                dict(
                    Name=str(proposal),
                    Description=f"Product data file",
                    Download=f'<a href="{rel_path}" download>CSV</a>',
                )
            )

            # Scaffold/chemistry CSV

            filename = f"Recipe_{proposal.hash}_chemistry"
            path = self.resource_dir / f"{filename}.csv"

            if not path.exists() or not self.skip_existing:
                proposal.write_chemistry_csv(path)

            rel_path = Path(self.resource_dir.name) / path.name

            table_data.append(
                dict(
                    Name=str(proposal),
                    Description=f"Chemistry review file",
                    Download=f'<a href="{rel_path}" download>CSV</a>',
                )
            )

        self.table(table_data)

    ### GRAPHS

    def funnel(
        self,
        log_y: bool = True,
        scaffolds: "CompoundSet | None" = None,
        num_inspirations: int | None = None,
        num_inspiration_sets: int | None = None,
    ) -> "plotly.graph_objects.Figure":

        if scaffolds is None:
            scaffolds = self.all_scaffolds
            scaffold_poses = self.all_scaffold_poses
            elabs = self.all_elabs
        else:
            scaffold_poses = scaffolds.poses
            elabs = scaffolds.elabs

        import plotly.express as px
        from numpy import log as np_log
        from pandas import DataFrame

        data = dict(
            number=[
                num_inspirations or scaffold_poses.num_inspirations,
                num_inspiration_sets or scaffold_poses.num_inspiration_sets,
                len(scaffolds),
                len(elabs),
                len(self.rgen.route_pool),
                len(self.proposal.products),
            ],
            category=[
                "Fragments",
                "Fragment Sets",
                "Scaffolds",
                "Elaborations",
                "Accessible Products",
                "Selected Products",
            ],
        )

        df = DataFrame(data)

        if log_y:
            y = "log_y"
            df["log_y"] = df.apply(lambda x: np_log(x["number"]), axis=1)
        else:
            y = "number"

        fig = px.funnel(df, x="category", y=y, text="number", log_y=False)

        fig.data[0].texttemplate = "%{text}"
        fig.update_layout(
            xaxis={"side": "top"},
        )

        fig.layout.xaxis.title.text = ""

        # title = title or f"<b>{animal.name}</b>: Reaction statistics"

        # if subtitle:
        #     title = f"{title}<br><sup><i>{subtitle}</i></sup>"

        # fig.update_layout(title=title, title_automargin=False, title_yref="container")

        return fig

    ### DUNDERS



import mcol

import mrich


class Target:
    """Object representing a protein target

    .. attention::

            :class:`.Target` objects should not be created directly. Instead use :meth:`.HIPPO.register_target` or :meth:`.Pose.target`.

    """

    _feature_cache = {}

    def __init__(
        self,
        db: "Database",
        id: int,
        name: str,
    ) -> None:

        self._db = db
        self._id = id
        self._name = name

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def id(self) -> int:
        """Returns the target's ID"""
        return self._id

    @property
    def name(self) -> str:
        """Returns the target's name"""
        return self._name

    @property
    def feature_ids(self) -> list[int]:
        """Returns the target's feature ID's"""
        feature_ids = self.db.select_where(
            query="feature_id",
            table="feature",
            key="target",
            value=self.id,
            none=False,
            multiple=True,
            sort="feature_chain_name, feature_residue_number",
        )

        if feature_ids:
            feature_ids = [v for v, in feature_ids]

        return feature_ids

    @property
    def features(self) -> list["Feature"]:
        """Returns the target's features"""
        if feature_ids := self.feature_ids:
            return [self.db.get_feature(id=i) for i in feature_ids]
        return None

    @property
    def subsites(self):

        from .subsite import Subsite

        records = self.db.select_where(
            table="subsite",
            key="target",
            value=self.id,
            multiple=True,
            query="subsite_id, subsite_name",
        )

        if not records:
            return []

        subsites = []
        for record in records:
            id, name = record
            subsite = Subsite(db=self.db, id=id, name=name, target_id=self.id)
            subsites.append(subsite)

        return subsites

    ### METHODS

    def calculate_features(
        self,
        protein: "mp.System",
        reference_id: int | None = None,
    ) -> list["Feature"]:
        """Calculate features from a protein system

        :param protein: `molparse.System` object, likely from :meth:`.Pose.protein_system`
        :returns: a list of :class:`.Feature` objects

        """

        if reference_id and reference_id in self._feature_cache:
            return self._feature_cache[reference_id]

        else:

            features = protein.get_protein_features()

            for f in features:
                self.db.insert_feature(
                    family=f.family,
                    target=self.id,
                    atom_names=[a.name for a in f.atoms],
                    residue_name=f.res_name,
                    residue_number=f.res_number,
                    chain_name=f.res_chain,
                    commit=False,
                )

            self.db.commit()

            features = self.features

            if reference_id:
                self._feature_cache[reference_id] = features

            return features

    ### DUNDERS

    def __str__(self) -> str:
        """Unformatted string representation"""
        return f'T{self.id} "{self.name}"'

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"



from .recipe import Recipe
from .cset import CompoundSet, IngredientSet

from tqdm import tqdm

from pathlib import Path
import json

from .tools import dt_hash

import mrich


class RRGMixin:
    """Mixin class for shared properties"""

    @property
    def db(self) -> "Database":
        """Get the linked HIPPO Database object"""
        return self._db

    @property
    def db_path(self) -> str:
        """Get the path of the linked Database"""
        return self._db_path

    @property
    def starting_recipe(self):
        """Get the starting recipe used in all generations"""
        return self._starting_recipe

    @property
    def suppliers_str(self) -> str:
        """SQL formatted tuple of suppliers"""
        return str(tuple(self.suppliers)).replace(",)", ")")

    @property
    def suppliers(self) -> list[str]:
        """List of suppliers"""
        return self._suppliers

    @property
    def max_lead_time(self) -> float:
        """Maximum lead-time constraint"""
        return self._max_lead_time

    @property
    def data_path(self):
        """File path for the JSON data export"""
        return self._data_path

    @property
    def recipe_dir(self):
        """File path for the JSON recipe export"""
        return self._recipe_dir

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        import mcol

        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __call__(self, *args, **kwargs):
        return self.generate(*args, **kwargs)

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"


class RandomRecipeGenerator(RRGMixin):
    """Class to create randomly sampled Recipe from a HIPPO Database"""

    def __init__(
        self,
        db,
        *,
        max_lead_time=None,
        # max_reactions = None,
        suppliers: list | None = None,
        start_with: Recipe | CompoundSet | IngredientSet = None,
    ):

        mrich.debug("RandomRecipeGenerator.__init__()")

        # Static parameters
        self._db_path = db.path
        self._max_lead_time = max_lead_time
        self._suppliers = suppliers
        self._starting_recipe = start_with

        mrich.var("database", self.db_path)
        mrich.var("max_lead_time", self.max_lead_time)
        mrich.var("suppliers", self.suppliers)

        # Database set up
        self._db = db

        # JSON I/O set up
        self._data_path = Path(str(self.db_path.name).replace(".sqlite", "_rgen.json"))
        if self.data_path.exists():
            mrich.warning(f"Will overwrite existing rgen data file: {self.data_path}")

        # Recipe I/O set up
        path = Path(str(self.db_path.name).replace(".sqlite", "_recipes"))
        mrich.writing(f"{path}/")
        path.mkdir(exist_ok=True)
        self._recipe_dir = path

        # Route pool
        mrich.debug("Solving route pool...")
        self._route_pool = self.get_route_pool()

        # dump data
        self.dump_data()

    ### FACTORIES

    @classmethod
    def from_json(cls, db: "Database", path: "Path | str"):
        """Construct the RandomRecipeGenerator from a JSON file

        :param db:
        :param path:

        """

        data = json.load(open(path, "rt"))

        self = cls.__new__(cls)

        self._db_path = Path(data["db_path"])
        self._recipe_dir = Path(data["recipe_dir"])
        self._max_lead_time = data["max_lead_time"]
        self._suppliers = data["suppliers"]

        self._starting_recipe = Recipe.from_json(
            db=db,
            path=None,
            data=data["starting_recipe"],
            allow_db_mismatch=True,
        )

        mrich.var("database", self.db_path)
        mrich.var("max_lead_time", self.max_lead_time)
        mrich.var("suppliers", self.suppliers)

        self._db = db

        # JSON I/O set up
        self._data_path = Path(path)

        # Route pool
        from .recipe import RouteSet

        self._route_pool = RouteSet.from_json(path=None, data=data["route_pool"], db=db)

        return self

    ### PROPERTIES

    @property
    def route_pool(self):
        """Get the RouteSet of all product reaction routes considered by this generator"""
        return self._route_pool

    ### POOL METHODS

    def get_route_pool(self, mini_test=False):
        """Construct the pool of routes that will be randomly sampled from

        :param mini_test:  (Default value = False)

        """

        """
			Explainer for SQL query:

			- get table of quoted compounds with a count of the valid suppliers
			- join routes, components, and the new table together and grouped by route count the unavailable reactants
			- return route ids where no reactants are unavailable

		"""

        if "route" not in self.db.table_names:
            mrich.error("route table not in Database")
            raise NotImplementedError

        assert self.suppliers_str
        if self.max_lead_time:
            raise NotImplementedError

        ### EXCLUDE PRODUCTS OF ROUTES IN STARTING RECIPE!!!

        route_ids = self.db.execute(
            f"""
		WITH possible_reactants AS (
			SELECT quote_compound, COUNT(CASE WHEN quote_supplier IN {self.suppliers_str} THEN 1 END) AS [count_valid] FROM quote
			GROUP BY quote_compound
		),

		route_reactants AS (
			SELECT route_id, route_product, 
            COUNT(
                CASE 
                    WHEN count_valid = 0 THEN 1 
                    WHEN count_valid IS NULL THEN 1 
                END) 
            AS [count_unavailable] FROM route
			INNER JOIN component ON component_route = route_id
			LEFT JOIN possible_reactants ON quote_compound = component_ref
			WHERE component_type = 2
			GROUP BY route_id
		)

		SELECT route_id FROM route_reactants
		WHERE count_unavailable = 0
		"""
        ).fetchall()

        if mini_test:
            route_ids = route_ids[:100]

        routes = [
            self.db.get_route(id=route_id)
            for route_id, in mrich.track(route_ids, prefix="Getting routes")
        ]

        from .recipe import RouteSet

        return RouteSet(self.db, routes)

    ### FILE I/O METHODS

    def dump_data(self):
        """ """

        data = {}

        data["db_path"] = str(self.db_path.resolve())
        data["recipe_dir"] = str(self.recipe_dir.resolve())
        data["max_lead_time"] = self.max_lead_time
        data["suppliers"] = self.suppliers
        data["starting_recipe"] = self.starting_recipe.get_dict(serialise_price=True)
        data["route_pool"] = self.route_pool.get_dict()

        mrich.writing(self.data_path)
        json.dump(data, open(self.data_path, "wt"), indent=4)

    def generate(
        self,
        budget: float = 10000,
        currency: str = "EUR",
        max_products=1000,
        max_reactions=1000,
        debug=False,
        max_iter=None,
        shuffle=True,
        balance_clusters: bool = False,
        permitted_clusters: None | set = None,
    ):
        """

        :param budget: float:  (Default value = 10000)
        :param currency: str:  (Default value = 'EUR')
        :param max_products:  (Default value = 1000)
        :param max_reactions:  (Default value = 1000)
        :param debug:  (Default value = True)
        :param max_iter:  (Default value = None)
        :param # pick_inner_cheapest:  (Default value = True)
        :param # add_size:  (Default value = 1)
        :param shuffle:  (Default value = True)

        """

        # construct filename

        out_file = self.recipe_dir / f"Recipe_{dt_hash()}.json"

        from .price import Price

        if not max_iter:
            max_iter = max_products + max_reactions

        budget = Price(budget, currency)

        recipe = self.starting_recipe.copy()

        recipe.reactants._supplier = self.suppliers

        # get the RouteSet
        pool = self.route_pool.copy()

        assert len(pool), "Route pool is empty!"

        if shuffle:
            mrich.debug("Shuffling Route pool")
            pool.shuffle()

        old_recipe = recipe.copy()

        mrich.var("route pool", len(pool))
        mrich.var("max_iter", max_iter)

        for i in mrich.track(range(max_iter), prefix="Generating Recipe..."):

            if debug:
                mrich.title(f"Iteration {i}")

            price = recipe.price
            mrich.set_progress_field("price", str(price))
            mrich.set_progress_field("#products", len(recipe.products))

            if debug:
                mrich.var("price", price)

            # pop a route
            if balance_clusters:
                candidate_route = pool.balanced_pop(
                    permitted_clusters=permitted_clusters
                )
            else:
                candidate_route = pool.pop()

            if debug:
                mrich.var("candidate_route", candidate_route)
            if debug:
                mrich.var("candidate_route.reactants", candidate_route.reactants.ids)

            if candidate_route.product in recipe.products:
                continue

            # add the route to the recipe
            if debug:
                mrich.var("#recipe.reactants", len(recipe.reactants))
            recipe += candidate_route
            if debug:
                mrich.var("#recipe.reactants", len(recipe.reactants))

            # calculate the new price
            try:
                new_price = recipe.price
            except AssertionError:
                mrich.error(
                    f"Something went wrong while calculating the price after adding {candidate_route=} to recipe"
                )
                raise

            if debug:
                mrich.var("new price", new_price)

            # Break if product pool depleted
            if not len(pool):
                stop_reason = "Product pool depleted"
                mrich.success(stop_reason)
                break

            # check breaking conditions
            if new_price > budget:
                recipe = old_recipe.copy()
                continue

            if len(recipe.reactions) > max_reactions:
                stop_reason = "Max #reactions exceeded"
                mrich.success(stop_reason)
                break

            if len(recipe.products) > max_products:
                stop_reason = "Max #products exceeded"
                mrich.success(stop_reason)
                break

            # accept change
            old_recipe = recipe.copy()

        else:
            stop_reason = "Max #iterations reached"
            mrich.warning(stop_reason)

        ### recalculate the products to see if any extra can be had for free?

        mrich.success(f"Completed after {i} iterations")

        metadict = {
            "rgen_data_path": str(self.data_path.resolve()),
            "rgen_db_path": str(self.db_path.resolve()),
            "rgen_recipe_dir": str(self.recipe_dir.resolve()),
            "rgen_max_lead_time": self.max_lead_time,
            "rgen_suppliers": self.suppliers,
            "gen_budget": budget.amount,
            "gen_currency": budget.currency,
            "gen_max_products": max_products,
            "gen_max_reactions": max_reactions,
            "gen_max_iter": max_iter,
            "gen_shuffle": shuffle,
            "gen_iterations": i,
            "gen_stop_reason": stop_reason,
            "gen_recipe_path": str(out_file.resolve()),
        }

        # write the Recipe JSON
        recipe.write_json(out_file, extra=metadict)

        return recipe

    ### DUNDERS

    def __str__(self) -> str:
        """Unformatted string representation"""
        return f"RandomRecipeGenerator(recipe_dir={self.recipe_dir})"


class RandomSelectionGenerator(RRGMixin):
    """Class to create randomly sampled (no-chemistry) Recipe from a HIPPO Database"""

    def __init__(
        self,
        db,
        *,
        # max_lead_time=None,
        suppliers: list | None = None,
        amount: float = 1.0,  # in mg
        start_with: Recipe | CompoundSet | IngredientSet = None,
        compounds: CompoundSet | None = None,
        quoted_only: bool = True,
    ):

        mrich.debug("RandomRecipeGenerator.__init__()")

        # Static parameters
        self._db_path = db.path
        # self._max_lead_time = max_lead_time
        self._suppliers = suppliers
        self._amount = amount
        self._quoted_only = quoted_only
        self._db = db
        # self._compound_pool = compounds

        mrich.var("database", self.db_path)
        # mrich.var("max_lead_time", self.max_lead_time)
        mrich.var("suppliers", self.suppliers)
        mrich.var("amount per compound", self.amount, unit="mg")
        mrich.var("quoted_only", self.quoted_only)

        self.get_starting_recipe(start_with)
        mrich.var("starting recipe", self.starting_recipe)

        # JSON I/O set up
        self._data_path = Path(str(self.db_path.name).replace(".sqlite", "_rgen.json"))
        if self.data_path.exists():
            mrich.warning(f"Will overwrite existing rgen data file: {self.data_path}")

        # Recipe I/O set up
        path = Path(str(self.db_path.name).replace(".sqlite", "_recipes"))
        mrich.writing(f"{path}/")
        path.mkdir(exist_ok=True)
        self._recipe_dir = path

        with mrich.spinner("Getting compound pool"):
            self.get_compound_pool(compounds)
        mrich.var("compound pool", self.compound_pool)

        # dump data
        self.dump_data()

    ### FACTORIES

    @classmethod
    def from_json(cls, db: "Database", path: "Path | str"):
        """Construct the RandomRecipeGenerator from a JSON file

        :param db:
        :param path:

        """

        data = json.load(open(path, "rt"))

        self = cls.__new__(cls)

        self._db_path = Path(data["db_path"])
        self._recipe_dir = Path(data["recipe_dir"])
        # self._max_lead_time = data["max_lead_time"]
        self._suppliers = data["suppliers"]
        self._amount = data["amount"]

        self._starting_recipe = Recipe.from_json(
            db=db,
            path=None,
            data=data["starting_recipe"],
            allow_db_mismatch=True,
        )

        mrich.var("database", self.db_path)
        mrich.var("suppliers", self.suppliers)
        mrich.var("amount", self.amount)
        mrich.var("starting_recipe", self.starting_recipe)

        self._db = db

        # JSON I/O set up
        self._data_path = Path(path)

        # Route pool
        self._compound_pool = IngredientSet.from_json(
            path=None, data=data["compound_pool"]["data"], db=db
        )
        mrich.var("compound_pool", self.compound_pool)

        return self

    ### PROPERTIES

    @property
    def amount(self) -> float:
        """Amount to quote each compound for"""
        return self._amount

    @property
    def quoted_only(self) -> bool:
        """Only consider compounds with quotes"""
        return self._quoted_only

    @property
    def compound_pool(self) -> "CompoundTable | CompoundSet":
        """The pool of compounds that will be chosen from"""
        return self._compound_pool

    ### METHODS

    def get_starting_recipe(
        self, start_with: "Recipe | CompoundSet | IngredientSet"
    ) -> Recipe:
        """Process start_with into Recipe object"""

        if isinstance(start_with, Recipe):
            if start_with.type != "NOCHEM":
                raise NotImplementedError("Only NOCHEM recipes are supported")
            self._starting_recipe = start_with
            return self._starting_recipe

        from .compound import Compound

        self._starting_recipe = Recipe(self.db)

        if start_with is not None:
            for item in start_with:
                if isinstance(item, Compound):
                    item = item.as_ingredient(amount=self._amount)

                self._starting_recipe.compounds.add(item)

        return self._starting_recipe

    def get_compound_pool(
        self, compounds: CompoundSet | None
    ) -> "CompoundTable | CompoundSet":

        if self.suppliers:
            raise NotImplementedError

        if compounds is None:

            # all compounds
            if not self.quoted_only:
                ids = self.db.select(
                    table="compound", query="compound_id", multiple=True
                )
                self._compound_pool = IngredientSet.from_compounds(
                    db=self.db, ids=[i for i, in ids], amount=self.amount
                )
                return self._compound_pool

            # get all compounds that have a quote

            sql = f"""
            SELECT quote_id, quote_compound, quote_amount, quote_supplier, MIN(quote_price) 
            FROM quote
            WHERE quote_amount >= {self.amount}
            GROUP BY quote_compound
            """

            records = self.db.execute(sql).fetchall()

            ingredients = [
                dict(
                    quote_id=i,
                    compound_id=c,
                    amount=self.amount,
                    quoted_amount=a,
                    supplier=None,
                    max_lead_time=None,
                )
                for i, c, a, s, p in records
            ]

            self._compound_pool = IngredientSet.from_ingredient_dicts(
                self.db, ingredients
            )

        else:

            # ignore quoting
            if not self.quoted_only:
                self._compound_pool = IngredientSet.from_compounds(
                    db=self.db, ids=compounds.ids, amount=self.amount
                )
                return self._compound_pool

            # get all compounds that have a quote

            sql = f"""
            SELECT quote_id, quote_compound, quote_amount, quote_supplier, MIN(quote_price) 
            FROM quote
            WHERE quote_amount >= {self.amount}
            AND quote_compound IN {compounds.str_ids}
            GROUP BY quote_compound
            """

            records = self.db.execute(sql).fetchall()

            ingredients = [
                dict(
                    quote_id=i,
                    compound_id=c,
                    amount=self.amount,
                    quoted_amount=a,
                    supplier=None,
                    max_lead_time=None,
                )
                for i, c, a, s, p in records
            ]

            self._compound_pool = IngredientSet.from_ingredient_dicts(
                self.db, ingredients
            )

    def dump_data(self):
        """ """

        data = {}

        data["db_path"] = str(self.db_path.resolve())
        data["recipe_dir"] = str(self.recipe_dir.resolve())
        # data["max_lead_time"] = self.max_lead_time
        data["amount"] = self.amount
        data["suppliers"] = self.suppliers
        data["starting_recipe"] = self.starting_recipe.get_dict(serialise_price=True)
        data["compound_pool"] = self.compound_pool.get_dict()

        mrich.writing(self.data_path)
        json.dump(data, open(self.data_path, "wt"), indent=4)

    def generate(
        self,
        budget: float = 10000,
        currency: str = "EUR",
        max_iter: int | None = None,
        max_compounds=1000,
        debug=False,
        shuffle=True,
        # balance_clusters: bool = False,
        # permitted_clusters: None | set = None,
    ):
        """ """

        # construct filename

        out_file = self.recipe_dir / f"Recipe_{dt_hash()}.json"

        from .price import Price

        budget = Price(budget, currency)

        recipe = self.starting_recipe.copy()

        recipe.compounds._supplier = self.suppliers

        # get the RouteSet
        pool = self.compound_pool.copy()

        assert len(pool), "Route pool is empty!"

        if shuffle:
            mrich.debug("Shuffling Route pool")
            pool.shuffle()

        old_recipe = recipe.copy()

        if not max_iter:
            max_iter = max_compounds * 3

        mrich.var("compound pool", pool)
        mrich.var("max_compounds", max_compounds)
        mrich.var("max_iter", max_iter)

        for i in mrich.track(range(max_iter), prefix="Generating Recipe..."):

            if debug:
                mrich.title(f"Iteration {i}")

            price = recipe.price
            mrich.set_progress_field("price", str(price))
            mrich.set_progress_field("#compounds", len(recipe.compounds))

            if debug:
                mrich.var("price", price)

            #     # pop a route
            #     if balance_clusters:
            #         candidate_route = pool.balanced_pop(
            #             permitted_clusters=permitted_clusters
            #         )
            #     else:
            candidate = pool.pop()

            if debug:
                mrich.var("candidate", candidate)

            if candidate in recipe.compounds:
                continue

            # add the route to the recipe
            recipe.compounds.add(candidate)

            # calculate the new price
            try:
                new_price = recipe.price
            except AssertionError:
                mrich.error(
                    f"Something went wrong while calculating the price after adding {candidate_route=} to recipe"
                )
                raise

            if debug:
                mrich.var("#compounds", recipe.num_compounds)
                mrich.var("new price", new_price)

            # Break if product pool depleted
            if not len(pool):
                stop_reason = "Compound pool depleted"
                mrich.success(stop_reason)
                break

            # check breaking conditions
            if new_price > budget:
                recipe = old_recipe.copy()
                continue

            if len(recipe.compounds) > max_compounds:
                stop_reason = "Max #compounds exceeded"
                mrich.success(stop_reason)
                break

            # accept change
            old_recipe = recipe.copy()

        else:
            stop_reason = "Max #iterations reached"
            mrich.warning(stop_reason)

        ### recalculate the products to see if any extra can be had for free?

        mrich.success(f"Completed after {i} iterations")

        metadict = {
            "rgen_data_path": str(self.data_path.resolve()),
            "rgen_db_path": str(self.db_path.resolve()),
            "rgen_recipe_dir": str(self.recipe_dir.resolve()),
            # "rgen_max_lead_time": self.max_lead_time,
            "rgen_suppliers": self.suppliers,
            "rgen_amount": self.amount,
            "gen_budget": budget.amount,
            "gen_currency": budget.currency,
            "gen_max_compounds": max_compounds,
            "gen_shuffle": shuffle,
            "gen_iterations": i,
            "gen_stop_reason": stop_reason,
            "gen_recipe_path": str(out_file.resolve()),
        }

        # write the Recipe JSON
        recipe.write_json(out_file, extra=metadict)

        return recipe

    ### DUNDERS

    def __str__(self) -> str:
        """Unformatted string representation"""
        return f"RandomSelectionGenerator(recipe_dir={self.recipe_dir})"



import pycule
from .quote import Quote
import requests

import logging

# logging.getLogger("requests").setLevel(logging.WARNING)
# logging.getLogger("mcule:core").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

ENAMINE_CATALOGUES = [
    "BB",
    "SCR",
    "MADE",
    "REAL",
]

ENAMINE_V2_CATALOGUES = {
    "building-blocks": "BB",
    "screening-compounds": "SCR",
    "made-building-blocks": "MADE",
    "real": "REAL",
    "bioactive-compounds": "BIO",
}

ENAMINE_V2_BASE_URL = "https://new.enaminestore.com/api/v2"

ENAMINE_V2_LEAD_TIME = {
    "BB": {
        "IN_STOCK": 5,
        "SYNTHESIS": 7,
    },
    "SCR": {
        "IN_STOCK": 5,
        "SYNTHESIS": 20,
    },
    "MADE": {
        "SYNTHESIS": 20,
    },
    "REAL": {
        "SYNTHESIS": 20,
    },
    "BIO": {
        "IN_STOCK": 5,
    },
}

import mrich


class Quoter:
    """Class to scrape catalogue data"""

    def __init__(
        self,
        supplier: str,
        username: str = None,
        password: str = None,
        token: str = None,
    ):
        """Currently 'Enamine' and 'MCule' are supported. Once initialised"""

        if supplier == "Enamine":

            self.supplier = "Enamine"
            self.batch_size = 2000
            self.catalogues = ENAMINE_CATALOGUES
            self._get_quote = self.get_enamine_quote
            self._get_batch_quote = self.get_enamine_batch_quote

            assert username
            assert password

            self.wrapper = pycule.core.EnamineWrapper(
                username=username, password=password
            )

        elif supplier == "MCule":

            self.supplier = "MCule"
            self.batch_size = 1000
            self._get_quote = self.get_mcule_quote
            self._get_batch_quote = self.get_mcule_batch_quote

            assert token

            self.wrapper = pycule.core.MCuleWrapper(authorisation_token=token)

        else:
            mrich.error(f'Unsupported supplier: "{supplier}"')
            raise NotImplementedError

    def get_quote(self, compound, **kwargs):
        """Get quotes for a compound

        :param compound:

        """
        return self._get_quote(compound, **kwargs)

    def get_batch_quote(self, compounds, **kwargs):
        """Get quotes for a compound set

        :param compounds:

        """
        return self._get_batch_quote(compounds, **kwargs)

    def __call__(self, compound):
        """Get quotes for a compound"""
        return self.get_quote(compound)

    ### ENAMINE

    def get_enamine_batch_quote(
        self,
        compounds,
        currency="USD",
        catalogues=None,
        exact=False,
        forms=False,
        analogues=False,
        equivalents=False,
    ):
        """

        :param compounds:
        :param currency:  (Default value = "USD")
        :param catalogues:  (Default value = None)
        :param exact:  (Default value = False)
        :param forms:  (Default value = False)
        :param analogues:  (Default value = False)
        :param equivalents:  (Default value = False)

        """

        import time
        from .cset import CompoundSet

        n_comps = len(compounds)
        mrich.var("batch size", n_comps)
        assert n_comps <= 2000

        unmatched = compounds.copy()
        matched = CompoundSet(compounds.db)
        errors = []

        # perform an exact search on the batch of smiles (for each catalog)

        payload = dict(compounds=compounds.smiles, currency=currency)

        include = []
        if forms:
            include.append("FORMS")
        if analogues:
            include.append("ANALOGS")
        if equivalents:
            include.append("EQUIVALENTS")
        if include:
            payload["include"] = include

        catalogues = catalogues or ENAMINE_V2_CATALOGUES

        t_start = time.perf_counter()

        # return payload

        for catalogue in catalogues:

            if exact:
                mrich.debug(
                    f"Exact search in Enamine:{ENAMINE_V2_CATALOGUES[catalogue]}..."
                )
            else:
                mrich.debug(
                    f"Similarity==1.0 search in Enamine:{ENAMINE_V2_CATALOGUES[catalogue]}..."
                )

            # send the request
            t_catalogue_start = time.perf_counter()
            if exact:
                url = (
                    ENAMINE_V2_BASE_URL
                    + f"/catalog/search/in/{catalogue}/by/SMILEs/EXACT"
                )
            else:
                url = (
                    ENAMINE_V2_BASE_URL
                    + f"/catalog/search/in/{catalogue}/by/SMILEs/SIMILARITY/1.00"
                )
            response = requests.post(url=url, json=payload)
            mrich.var(
                f"{catalogue} request time = ", time.perf_counter() - t_catalogue_start
            )

            # process the request
            if (status := response.status_code) != 200:
                mrich.error(f"Request error: status={status}")
                errors.append(
                    dict(catalogue=catalogue, status=status, response=response)
                )
                continue

            data = response.json()

            results = data["results"]
            products = results["products"]

            # register the quotes in the database
            this_matched = self.parse_enamine_bulk_products(compounds.db, products)

            if this_matched:
                mrich.success(f"Added quotes for {len(this_matched)} compounds")

            matched += this_matched
            unmatched -= this_matched

            return data

            # break

        mrich.var("Total time = ", time.perf_counter() - t_start)

        if unmatched:
            mrich.error(f"Did not find quotes for {len(unmatched)} compounds")

        if errors:
            mrich.error(f"{len(errors)} failed requests")

        return dict(matched=matched, unmatched=unmatched, errors=errors)

    def parse_enamine_bulk_products(self, db, products):
        """

        :param db:
        :param products:

        """

        from .tools import flat_inchikey
        from .cset import CompoundSet

        matched = CompoundSet(db)

        for product in products:

            prices = product["prices"]
            product = product["product"]

            # quote-level data from product
            entry = product["code"]
            smiles = product["smile"]
            currency = prices["currency"]
            catalogue = ENAMINE_V2_CATALOGUES[product["catalog"]]

            # check if the product has a name
            if not entry:
                mrich.warning(f"Product in {catalogue} but no entry-name. {smiles=}")

            if p := product["purity"]:
                purity = product["purity"] / 100
            else:
                purity = None

            assert catalogue, product
            assert currency, product

            # identify the compound
            compound_id = db.get_compound_id(inchikey=flat_inchikey(smiles))
            matched.add(compound_id)

            # loop through prices
            for pack in prices["all"]:

                assert pack["weight"]["measure"] == "mg"
                amount = pack["weight"]["amount"]
                price = pack["price"]

                availability_str = pack["weight"]["available"]

                try:
                    lead_time = ENAMINE_V2_LEAD_TIME[catalogue][availability_str]
                except KeyError:
                    mrich.error(
                        f"Unsupported {availability_str=} [{compound_id=}, {entry=}]"
                    )
                    print(pack)
                    continue

                # register the quote
                quote_id = db.insert_quote(
                    compound=compound_id,
                    supplier="Enamine",
                    catalogue=catalogue,
                    entry=entry,
                    amount=amount,
                    price=price,
                    currency=currency,
                    purity=purity,
                    lead_time=lead_time,
                    smiles=smiles,
                    commit=False,
                )

            # break

        mrich.debug("Committing changes to the database...")
        db.commit()

        return matched

    def get_enamine_quote(self, compound, currency="USD"):
        """

        :param compound:
        :param currency:  (Default value = "USD")

        """

        assert compound._table == "compound"

        try:

            smiles = compound.smiles

            mrich.header(f"Exact search: {smiles=}")

            for catalogue in self.catalogues:

                mrich.header(f"Searching in {self.supplier} {catalogue}...")

                try:
                    result = self.wrapper.exactsearch(
                        smiles=compound.smiles, currency=currency, catalogue=catalogue
                    )
                except requests.ConnectionError as e:
                    mrich.error(f"ConnectionError: {e}")
                    return None

                # try again but with a similarity search
                if result["response"]["result"]["code"] != 0:
                    try:
                        result = self.wrapper.similaritysearch(
                            smiles, similarity_value=1.0, catalogue=catalogue
                        )
                    except requests.ConnectionError as e:
                        mrich.error(f"ConnectionError: {e}")
                        return None

                if result["response"]["result"]["code"] != 0:
                    continue

                data = result["response"]["data"]

                if len(data) < 1:
                    print(result["response"])
                    raise NotImplementedError

                data = data[0]

                name = data["Id"]

                mrich.success(f"Found in {catalogue} w/ id={name}")

                break

            else:

                mrich.error("No match found")
                return None

            ### get the information

            mrich.header(f"Searching by ID in {self.supplier} {catalogue}...")

            for catalogue in self.catalogues:

                result = self.wrapper.compoundidsearch(name, catalogue=catalogue)

                if result["response"]["result"]["code"] == 0:
                    try:
                        mrich.success(f"Found in {catalogue}")
                        return (
                            self.parse_enamine_response(result, compound, catalogue),
                            result,
                        )
                    except NoDataInReponse:
                        pass

            else:

                mrich.error("No catalog entry")
                return None

        except KeyboardInterrupt:
            mrich.warning("Interrupted quoting")

    def pick_enamine_exact_data(self, smiles, data):
        """

        :param smiles:
        :param data:

        """

        if len(data) == 1:
            return data[0]["Id"]

        if len(data) == 0:
            raise NoDataInResponse(smiles)

        # sort the results by increasing lead_time, and increasing stereo-complexity
        data = sorted(
            data,
            key=lambda x: (
                self.parse_enamine_delivery_string(x["deliveryDays"]),
                stereo_complexity(x["smile"]),
            ),
        )
        mout.warning(
            f'Picking quickest and least stereo complex result from: {[x["Id"] for x in data]}'
        )
        return data[0]["Id"]

    def parse_enamine_response(self, result, compound, catalogue):
        """

        :param result:
        :param compound:
        :param catalogue:

        """

        quotes = []

        for i, data in enumerate(result["response"]["data"]):

            if i == 0:

                # update compound metadata
                metadata = dict()
                metadata["enamine_id"] = data["Id"]
                metadata["alias"] = data["name"]
                metadata["cas"] = data["cas"]
                metadata["mfcd"] = data["mfcd"]
                metadata["formula"] = data["formula"]
                metadata["molecular_weight"] = data["mw"]
                compound.metadata.update(metadata)

                """
				do something with: 
				* availability
				* storageCond
				* productUrl
				* lastUpdate
				"""

            lead_time = self.parse_enamine_delivery_string(
                delivery := data["deliveryDays"]
            )
            entry = data["Id"]
            smiles = data["smile"]

            if purity := data["purity"]:
                purity /= 100

            # if not (packs := data['packs']):

            if not (packs := data["packs"]) and "synthesis" in delivery:

                mrich.warning("Hardcoded REAL quote")

                compound.db.insert_quote(
                    compound=compound,
                    supplier="Enamine",
                    catalogue=catalogue,
                    entry=entry,
                    amount=20,
                    price=330,
                    currency="USD",
                    purity=None,
                    lead_time=15,
                    smiles=smiles,
                )

            else:
                for pack in packs:
                    self.parse_enamine_pack(
                        compound, entry, purity, catalogue, pack, lead_time, smiles
                    )

    def pick_enamine_data(self, comp_id, data):
        """

        :param comp_id:
        :param data:

        """

        data = [d for d in data if d["Id"] == comp_id]

        if len(data) == 0:
            raise NoDataInReponse

        elif len(data) == 1:
            return data[0]

        data = [d for d in data if not any([p["price"] == 0.0 for p in d["packs"]])]

        if len(data) > 1:
            mrich.warning(
                f"Taking first data entry after stripping non-priced ({comp_id})"
            )
        return data[0]

    def parse_enamine_delivery_string(self, string):
        """

        :param string:

        """

        if string.startswith("regular delivery, ") and string.endswith("days"):
            return int(string.removeprefix("regular delivery, ").removesuffix(" days"))

        if string.startswith("backorder, ") and string.endswith("days"):
            return int(string.removeprefix("backorder, ").removesuffix(" days"))

        if string == "3 weeks synthesis time":
            return 15

        raise Exception(f"Unexpected delivery string: {string}")

    def parse_enamine_pack(
        self, compound, entry, purity, catalogue, pack, lead_time, smiles
    ):
        """

        :param compound:
        :param entry:
        :param purity:
        :param catalogue:
        :param pack:
        :param lead_time:
        :param smiles:

        """

        supplier = "Enamine"

        match pack["measure"]:
            case "g":
                amount = pack["amount"] * 1000
            case "mg":
                amount = pack["amount"]
            case _:
                raise NotImplementedError(
                    f'Not supported pack measure: {pack["measure"]}'
                )

        price = pack["price"]
        currency = pack["currencyName"]

        if pack["status"] not in ["Normal", "Ice pack", ""]:
            print(pack)
            mrich.warning(f"{pack['status']=}")

        if not price:
            mrich.warning(f"Skipping price-less Enamine pack ({entry})")
            return None

        compound.db.insert_quote(
            compound=compound,
            supplier=supplier,
            catalogue=catalogue,
            entry=entry,
            amount=amount,
            price=price,
            currency=currency,
            purity=purity,
            lead_time=lead_time,
            smiles=smiles,
        )

    ### MCULE

    def get_mcule_quote(self, compound, exact=False):
        """

        :param compound:
        :param exact:  (Default value = False)

        """

        try:

            # single exact query

            smiles = compound.smiles

            if exact:
                mrich.header(f"Exact search: {smiles=}")
                result = self.wrapper.singlequerysearch(smiles)
            else:
                mrich.header(f"Similarity==1.0 search: {smiles=}")
                result = self.wrapper.similaritysearch(smiles, threshold=1.0)

            if result["response"]["results"]:

                results = result["response"]["results"]

                mrich.header(results)

                mcule_ids = [data["mcule_id"] for data in results]
                smiles = [data["smiles"] for data in results]

                mrich.success(f"Found w/ ids={mcule_ids}")

                for mcule_id, smile in zip(mcule_ids, smiles):

                    # mrich.header(f'{self.wrapper.compoundavailability(mcule_id)=}')
                    # mrich.header(f'{self.wrapper.compounddetails(mcule_id)=}')
                    result = self.wrapper.compoundprices(mcule_id)

                    for pack in result["response"]["best_prices"]:
                        self.parse_mcule_pack(
                            compound.db, compound, mcule_id, pack, smile
                        )

                    # mrich.header(f'{self.wrapper.compoundpricesamount(mcule_id)=}')

            return result

        except KeyboardInterrupt:
            mrich.warning("Interrupted quoting")

    def get_mcule_batch_quote(self, compounds):
        """

        :param compounds:

        """

        import time
        from .cset import CompoundSet
        from .tools import flat_inchikey

        db = compounds.db

        n_comps = len(compounds)

        matched = CompoundSet(db)

        mrich.var("batch size", n_comps)
        assert n_comps <= 1000

        # bulk query compound availability

        t_start = time.perf_counter()
        mrich.title("MCule bulk availability query...")
        data = self.wrapper.multiplequerieswithavailability(compounds.smiles)

        mrich.var("availability query time", time.perf_counter() - t_start)

        results = data["response"]["results"]

        mrich.var("#results", len(results))

        mrich.title("MCule single price queries...")
        for result in mrich.track(results, prefix="Querying MCule"):

            entry = result["mcule_id"]
            smiles = result["smiles"]

            # match to compound
            compound_id = db.get_compound_id(inchikey=flat_inchikey(smiles))

            if not compound_id:
                mrich.error(f"Could not find compound matching {smiles=}")
                continue

            matched.add(compound_id)

            prices = self.wrapper.compoundprices(entry)

            try:
                best_prices = prices["response"]["best_prices"]
            except KeyError as e:
                print(prices)
                mrich.error(f"Unsupported prices {e}")
                continue

            for pack in prices["response"]["best_prices"]:
                try:
                    self.parse_mcule_pack(
                        db, compound_id, entry, pack, smiles, commit=False
                    )
                except KeyError as e:
                    print(pack)
                    mrich.error(f"Unsupported pack {e}")
                    continue

        mrich.var("Total time = ", time.perf_counter() - t_start)

        unmatched = compounds - matched

        if unmatched:
            mrich.error(f"Did not find quotes for {len(unmatched)} compounds")

        mrich.success("Committing changes to the database...")
        db.commit()

        return dict(matched=matched, unmatched=unmatched)

    def parse_mcule_pack(self, db, compound, entry, pack, smiles, commit=True):
        """

        :param db:
        :param compound:
        :param entry:
        :param pack:
        :param smiles:
        :param commit:  (Default value = True)

        """

        supplier = "MCule"

        match pack["unit"]:
            case "g":
                amount = pack["amount"] * 1000
            case "mg":
                amount = pack["amount"]
            case _:
                raise NotImplementedError(f'Not supported pack measure: {pack["unit"]}')

        price = pack["price"]
        currency = pack["currency"]

        if "purity" in pack:
            purity = pack["purity"] / 100
        else:
            purity = None

        lead_time = pack["delivery_time_working_days"]

        db.insert_quote(
            compound=compound,
            supplier=supplier,
            catalogue=None,
            entry=entry,
            amount=amount,
            price=price,
            currency=currency,
            purity=purity,
            lead_time=lead_time,
            smiles=smiles,
            commit=commit,
        )


class NoDataInReponse(Exception):
    """ """

    ...


class NotInCatalogues(Exception):
    """ """

    ...



import mcol


class Subsite:
    """
    Class representing a subsite/subsite on a protein :class:`.Target`

    .. attention::

        :class:`.Subsite` objects should not be created directly. Instead use :meth:`.Target.subsites`.

    """

    _table = "subsite"

    def __init__(self, db: "Database", id: int, target_id: int, name: str):

        self._db = db
        self._id = id
        self._target_id = target_id
        self._name = name
        self._metadata = None
        self._target = None

    ### FACTORIES

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def id(self) -> int:
        """Returns the SubsiteTag's database ID"""
        return self._id

    @property
    def table(self):
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def target(self) -> "Target":
        """Returns the associated protein :class:`.Target`"""
        if self._target is None:
            self._target = self.db.get_target(id=self.target_id)
        return self._target

    @property
    def target_id(self) -> int:
        """Returns the associated protein :class:`.Target` ID"""
        return self._target_id

    @property
    def name(self) -> str:
        """The subsite name"""
        return self._name

    @property
    def metadata(self) -> "MetaData":
        """Returns the SubsiteTag's metadata"""
        if self._metadata is None:
            self._metadata = self.db.get_metadata(table="subsite", id=self.id)
        return self._metadata

    @property
    def poses(self) -> "PoseSet | None":
        """Return all poses in this subsite"""
        from .pset import PoseSet

        indices = self.db.select_where(
            table="subsite_tag",
            query="subsite_tag_pose",
            multiple=True,
            key="ref",
            value=self.id,
        )
        indices = [i for i, in indices]
        if not indices:
            return None
        return PoseSet(self.db, indices, name="poses in {self}")

    ### METHODS

    ### DUNDERS

    def __str__(self):
        """Unformatted string representation"""
        return f"S{self.id}: {self.target.name}->{self.name}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"


class SubsiteTag:
    """
    Class representing a tag assigning a :class:`.Subsite` to a :class:`.Pose`

    .. attention::

        :class:`.SubsiteTag` objects should not be created directly. Instead use :meth:`.Pose.subsites`.

    """

    _table = "subsite_tag"

    def __init__(self, db: "Database", id: int, subsite_id: int, pose_id: int):

        self._db = db
        self._id = id
        self._subsite_id = subsite_id
        self._pose_id = pose_id
        self._metadata = None

        self._name = db.get_subsite_name(id=subsite_id)

    ### FACTORIES

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def id(self) -> int:
        """Returns the SubsiteTag's database ID"""
        return self._id

    @property
    def table(self):
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def pose_id(self) -> int:
        """Returns the associated :class:`.Pose`'s database ID"""
        return self._pose_id

    @property
    def subsite_id(self):
        """Returns the associated :class:`.Subsite`'s database ID"""
        return self._subsite_id

    @property
    def name(self):
        """Returns the associated :class:`.Subsite`'s name"""
        return self._name

    @property
    def metadata(self) -> "MetaData":
        """Returns the SubsiteTag's metadata"""
        if self._metadata is None:
            self._metadata = self.db.get_metadata(table="subsite_tag", id=self.id)
        return self._metadata

    ### METHODS

    ### DUNDERS

    def __str__(self):
        """Unformatted string representation"""
        return self.name

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f'{mcol.bold}{mcol.underline}"{self}"{mcol.unbold}{mcol.ununderline}'

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f'[bold underline]"{self}"'



import mrich
from mrich import print
import molparse as mp

from molparse.rdkit.features import FEATURE_FAMILIES, INTERACTION_TYPES

INTERACTION_TYPES = list(INTERACTION_TYPES.values()) + ["VdWContact"]
FEATURE_FAMILIES = list(FEATURE_FAMILIES) + ["VdWSphere"]


def guess_feature_families(
    interaction_type: str, prot_atom_names: list[str], lig_atom_ids: list[int]
) -> tuple[str, str, str]:
    """Guess feature families from atom names

    :param interaction_type: interaction type

    """

    match interaction_type:
        case "VdWContact":
            lig_feature_family = "VdWSphere"
            prot_feature_family = "VdWSphere"

        case "Hydrophobic":
            lig_feature_family = (
                "LumpedHydrophobe" if len(lig_atom_ids) > 1 else "Hydrophobe"
            )
            prot_feature_family = (
                "LumpedHydrophobe" if len(prot_atom_names) > 1 else "Hydrophobe"
            )

        case "Anionic":
            lig_feature_family = "NegIonizable"
            prot_feature_family = "PosIonizable"
            interaction_type = "Electrostatic"

        case "Cationic":
            lig_feature_family = "PosIonizable"
            prot_feature_family = "NegIonizable"
            interaction_type = "Electrostatic"

        case "CationPi":
            lig_feature_family = "PosIonizable"
            prot_feature_family = "Aromatic"
            interaction_type = "π-cation"

        case "PiCation":
            lig_feature_family = "Aromatic"
            prot_feature_family = "PosIonizable"
            interaction_type = "π-cation"

        case "PiStacking":
            lig_feature_family = "Aromatic"
            prot_feature_family = "Aromatic"
            interaction_type = "π-stacking"

        case "PiStacking":
            lig_feature_family = "Aromatic"
            prot_feature_family = "Aromatic"
            interaction_type = "π-stacking"

        case "EdgeToFace":
            lig_feature_family = "Aromatic"
            prot_feature_family = "Aromatic"
            interaction_type = "π-stacking (EdgeToFace)"

        case "FaceToFace":
            lig_feature_family = "Aromatic"
            prot_feature_family = "Aromatic"
            interaction_type = "π-stacking (FaceToFace)"

        case "HBAcceptor":
            lig_feature_family = "Acceptor"
            prot_feature_family = "Donor"
            interaction_type = "Hydrogen Bond"

        case "HBDonor":
            lig_feature_family = "Donor"
            prot_feature_family = "Acceptor"
            interaction_type = "Hydrogen Bond"

        case "XBAcceptor":
            lig_feature_family = "Acceptor"
            prot_feature_family = "Donor"
            interaction_type = "Halogen Bond"

        case "XBDonor":
            lig_feature_family = "Donor"
            prot_feature_family = "Acceptor"
            interaction_type = "Halogen Bond"

        case "MetalAcceptor":
            lig_feature_family = None
            prot_feature_family = "Metal"
            interaction_type = "Metal complexation"

        case "MetalDonor":
            lig_feature_family = "Metal"
            prot_feature_family = None
            interaction_type = "Metal complexation"

        case _:
            prot_feature_family = None
            lig_feature_family = None

            mrich.error("Unsupported interaction_type", interaction_type)

    assert interaction_type in INTERACTION_TYPES
    assert prot_feature_family in FEATURE_FAMILIES or prot_feature_family is None
    assert lig_feature_family in FEATURE_FAMILIES or lig_feature_family is None

    return prot_feature_family, lig_feature_family, interaction_type


def parse_prolif_interactions(
    pose, fp, protonated_sys, table: str = "temp_interaction", debug: bool = False
):

    target_id = pose.target.id

    for key, value in fp.ifp[0].items():

        res_number = key[1].number
        res_name = key[1].name
        chain_name = key[1].chain

        chain = protonated_sys.get_chain(chain_name)
        residue = chain.residues.get_matches(number=res_number)

        prot_mol = residue.rdkit_mol
        prot_group = mp.AtomGroup.from_pdb_block(mp.rdkit.mol_to_pdb_block(prot_mol))

        if not residue.name == res_name:
            mrich.debug(protonated_sys.name + ".pdb")
            raise AssertionError(
                f"Residue name mismatch: [sys]={residue.name} [prolif]={res_name}"
            )

        for interaction_type, interaction_dicts in value.items():

            for interaction_dict in interaction_dicts:

                angle = interaction_dict.get("angle")
                distance = interaction_dict.get("distance")

                lig_atom_ids = list(interaction_dict["indices"]["ligand"])

                # insert a dummy protein feature
                prot_interaction_atoms = [
                    prot_group.atoms[i] for i in interaction_dict["indices"]["protein"]
                ]
                prot_atom_names = [a.name for a in prot_interaction_atoms]

                # rename stuff
                prot_family, lig_family, interaction_type = guess_feature_families(
                    interaction_type, prot_atom_names, lig_atom_ids
                )

                feature_id = pose.db.insert_feature(
                    family=prot_family,
                    target=target_id,
                    chain_name=chain_name,
                    residue_name=res_name,
                    residue_number=residue.number,
                    atom_names=prot_atom_names,
                    commit=False,
                )

                if not feature_id:

                    sql = f"""
                    feature_target = {target_id} 
                    AND feature_family = '{prot_family}' 
                    AND feature_chain_name = '{chain_name}' 
                    AND feature_residue_name = '{res_name}' 
                    AND feature_residue_number = {residue.number}
                    AND feature_atom_names = '{" ".join(sorted(prot_atom_names))}'
                    """

                    try:
                        (feature_id,) = pose.db.select_id_where(
                            table="feature", key=sql
                        )
                    except:
                        feature_id = pose.db.insert_feature(
                            family=prot_family,
                            target=target_id,
                            chain_name=chain_name,
                            residue_name=res_name,
                            residue_number=residue.number,
                            atom_names=prot_atom_names,
                            warn_duplicate=True,
                            commit=False,
                        )
                        raise

                # insert into the Database
                interaction_id = pose.db.insert_interaction(
                    feature=feature_id,
                    pose=pose.id,
                    type=interaction_type,
                    family=lig_family,
                    atom_ids=[i + 1 for i in lig_atom_ids],
                    prot_coord=None,
                    lig_coord=None,
                    distance=distance,
                    angle=angle,
                    energy=None,
                    commit=False,
                    table=table,
                )

                if debug:
                    mrich.debug(
                        f"Residue: {res_name} {res_number} {chain_name}. Interaction: {interaction_type}. Ligand: {lig_family} ({lig_atom_ids}). Protein: {prot_family} ({prot_atom_names})"
                    )
